"""
This module defines interfaces for each tool or decision maker used in the scraping process.
This makes it easy to test new ways or to tune one aspect of the scraper while keeping most of the code unchanged.

See the :py:mod:`swisstext.cmd.scraping.tools` module for implementations.
"""
from typing import List, Union


class IHtmlConverter:
    """
    [ABSTRACT] This tool is in charge of extracting text from HTML.
    """

    def to_text(self, html: Union[str, bytes], encoding=None, **kwargs) -> str:
        """[ABSTRACT] Should extract the text from HTML."""
        return ''


class INormalizer:
    """
    A normalizer should take a [long] text (extracted from a web page) and clean it in a consistant way.
    For example: uncurling quotes, normalizing punctuation, fix unicode, etc.
    The default implementation just return the text as-is.
    """

    def normalize(self, text: str, **kwargs) -> str:
        """This should be overriden. The default implementation just returns the text as-is."""
        return text

    def normalize_all(self, texts: List[str], **kwargs) -> List[str]:
        """Calls normalize on each element of text."""
        return [self.normalize(t, **kwargs) for t in texts]


class ISplitter:
    """
    A splitter should take a [long] text (extracted from a web page) and split it into well-formed sentences.
    The default implementation just splits on the newline character.
    """

    def split(self, text: str, **kwargs) -> List[str]:
        """This should be overriden. The default implementation just splits on newlines."""
        return text.splitlines()

    def split_all(self, texts: List[str], **kwargs) -> List[str]:
        """Takes a list of texts and returns a list of sentences (see :py:meth:`split`)."""
        return [splitted for t in texts for splitted in self.split(t, **kwargs)]


class IFilterer:
    """
    A sentence filter should be able to tell if a given sentence is well-formed (i.e. valid) or not.
    """

    def is_valid(self, sentence: str, **kwargs) -> bool:
        """This should be overriden. The default implementation just returns true."""
        return True

    def filter(self, sentences: List[str], **kwargs) -> List[str]:
        """Filter a list of sentences by calling :py:meth:`ISentenceFilter.is_valid` on each element."""
        return [s for s in sentences if self.is_valid(s, **kwargs)]



#!/usr/bin/env python3

"""
Reimplementation of moses' split-sentences.perl in pure Python 3. The behavior is equivalent to
https://bitbucket.org/luismsgomes/mosestokenizer/src/default/src/mosestokenizer/split-sentences.perl.

Changed compared to the latest split-sentences.perl:

* no support for Chinese, Hindi and Gujarati (those if code blocks were simply ignored)
* addition of a "more" option that splits on ":;" characters
* split returns a list of sentences instead of a string
* '<P>' are not added when encountering a "closing" empty line in the input (-> all empty lines are just ignored)

.. note::

    A better implementation (at least for swisstext) is
    :py:class:`~swisstext.cmd.scraping.tools.mocy_splitter.MocySplitter`

"""

import os
import re
import sys
import regex

from ..interfaces import ISplitter


# Perl Regex substitutions:
# * \p{IsPi} => \p{Pi} or \p{Initial_Punctuation}: any kind of opening quote.
# * \p{IsPf} => \p{Pf} or \p{Final_Punctuation}: any kind of closing quote.


class MosesSplitter(ISplitter):
    """Python implementation of Moses' split_sentences.perl"""

    def __init__(self, lang='de', prefix_file=None, more=True):
        """
        :param lang: nonbreaking_prefix file to load (available: en, de)
        :param prefix_file: path to a custom nonbreaking_prefix file
        :param more: if set, systematically split on :;
        """
        self.lang = lang
        self.more = more
        self.nb_prefixes = self.load_nb_prefixes(lang, prefix_file)

    def split(self, input_text):  # -> List[str]
        # equivalent of process_text in moses, but returns a list
        current_paragraph = ''
        splits = []
        for line in input_text.split('\n'):
            if not line or line.isspace() or (line.startswith('<') and line.endswith('>')):
                # Time to process this block; we've hit a blank or <p>
                if current_paragraph:
                    splits.extend(self._do_it_for(current_paragraph, line))
                    # don't happend '<P>' on empty lines
                    # if not len(line) or line.isspace() and current_paragraph:
                    #     splits.append('<P>')  ## If we have text followed by <P>
                    current_paragraph = ""
            else:
                current_paragraph += line + ' '

        if current_paragraph:
            # Do the leftover text.
            splits.extend(self._do_it_for(current_paragraph))

        return splits

    def _do_it_for(self, input_text, markup=''):
        # process one paragraph
        if not input_text:
            return ''
        text = self.split_paragraph(input_text, self.nb_prefixes, self.more)
        if markup.startswith('<') and markup.endswith('>'):
            return f'{text}\n{markup}'
        return text

    @classmethod
    def cleanup_spaces(cls, text):
        """Normalize spaces in a text."""
        # clean up spaces
        text = re.sub(' +', ' ', text)
        text = re.sub('\n ', '\n', text)
        text = re.sub(' \n ', '\n', text)
        return text.strip()

    @classmethod
    def split_paragraph(cls, text, nb_prefixes, more=False):  # -> List[str]
        """
        Handle on paragraph of text.

        :param text: the paragraph to split
        :param nb_prefixes: the dictionary of nonbreaking_prefix (see perl implementation/doc)
        :param more: if set, systematically split on :;
        :return: a list of sentences
        """
        # text is one paragraph, this is equivalent to the preprocess perl method.

        # clean up spaces
        text = cls.cleanup_spaces(text)

        ##### Add sentence breaks as needed #####
        if more:
            # this one is present in the python wrapper, see
            # https://bitbucket.org/luismsgomes/mosestokenizer/src/default/src/mosestokenizer/split-sentences.perl
            text = regex.sub(r'([\:;])', r'\1\n', text)

        # Non-period end of sentence markers (?!) followed by sentence starters.
        text = regex.sub(r'([?!]) +([\'\"\(\[\¿\¡\p{Pi}]*[\p{IsUpper}])', r'\1\n\2', text)

        # Multi-dots followed by sentence starters.
        text = regex.sub(r'(\.[\.]+) +([\'\"\(\[\¿\¡\p{Pi}]*[\p{IsUpper}])', r'\1\n\2', text)

        # Add breaks for sentences that end with some sort of punctuation
        # inside a quote or parenthetical and are followed by a possible
        # sentence starter punctuation and upper case.
        text = regex.sub(r'([?!\.][\ ]*[\'\"\)\]\p{Pf}]+) +([\'\"\(\[\¿\¡\p{Pi}]*[\ ]*[\p{IsUpper}])', r'\1\n\2', text)

        # Add breaks for sentences that end with some sort of punctuation,
        # and are followed by a sentence starter punctuation and upper case.
        text = regex.sub(r'([?!\.]) +([\'\"\(\[\¿\¡\p{Pi}]+[\ ]*[\p{IsUpper}])', r'\1\n\2', text)

        # Special punctuation cases are covered. Check all remaining periods.
        words = text.split(' ')
        text = ''
        for i in range(len(words) - 1):
            m = regex.search(r'([\p{IsAlnum}\.\-]*)([\'\"\)\]\%\p{Pf}]*)(\.+)$', words[i])
            if m is not None:
                # Check if $1 is a known honorific and $2 is empty, never break.
                prefix, starting_punct, _ = m.groups()
                if prefix and nb_prefixes.get(prefix, -1) == 1 and not starting_punct:
                    pass  # Not breaking;
                elif regex.search(r'(\.)[\p{IsUpper}\-]+(\.+)$', words[i]) is not None:
                    pass  # Not breaking - upper case acronym
                elif regex.search(r'^([ ]*[\'\"\(\[\¿\¡\p{Pi}]*[ ]*[\p{IsUpper}0-9])', words[i + 1]):
                    # The next word has a bunch of initial quotes, maybe a
                    # space, then either upper case or a number
                    if not (prefix and nb_prefixes.get(prefix, -1) == 2 and not starting_punct
                            and regex.search('^[0-9]+', words[i + 1])):
                        # We always add a return for these, unless we have a
                        # numeric non-breaker and a number start.
                        words[i] = words[i] + '\n'
            text += words[i] + ' '

        # We stopped one token from the end to allow for easy look-ahead.
        # Append it now.
        text = text + words[-1]

        # clean up spaces
        text = cls.cleanup_spaces(text)

        return text.split('\n')

    @classmethod
    def load_nb_prefixes(cls, lang='en', prefix_file=None):
        """
        Read the nonbreaking_prefixes from a file

        :param lang: the language to load
        :param prefix_file: a custom file to load (has priority over lang)
        :return: the nonbreaking_prefix dictionary, with key=prefix and value=1|2.
                 1 means apply anywhere, 2 means only applies when followed by digits.
        """
        if prefix_file is None:
            this_file = os.path.realpath(__file__)
            prefix_file = this_file.replace('_orig.py', f'_prefixes.{lang}.txt')  # TODO

            if not os.path.isfile(prefix_file):
                print(
                    f'WARNING: No known abbreviations for language {lang}, attempting fall-back to English version...',
                    file=sys.stderr)
                prefix_file = prefix_file.replace(f'.{lang}.txt', '.en.txt')
        else:
            print(f'Using custom prefix file: {prefix_file}', file=sys.stderr)

        prefixes = dict()
        with open(prefix_file) as f:
            for line in f:
                line = line.strip()
                if len(line) == 0 or line.startswith('#'):
                    continue

                if line.endswith('#NUMERIC_ONLY#'):
                    prefixes[line.replace('#NUMERIC_ONLY#', '').strip()] = 2
                else:
                    prefixes[line] = 1
        return prefixes


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', type=argparse.FileType('r'), default='-')
    parser.add_argument('-o', '--out', type=argparse.FileType('w'), default='-')
    parser.add_argument('-l', '--lang', default='de')
    parser.add_argument('-pf', '--prefix-file', default=None)
    parser.add_argument('-m', '--more', default=False, action='store_true')

    args = parser.parse_args()
    splitter = MosesSplitter(
        lang=args.lang,
        prefix_file=args.prefix_file,
        more=args.more)

    args.out.write('\n'.join(
        splitter.split(args.input.read())
    ))



import re
from typing import Union

from bs4 import BeautifulSoup

from phrasal.interfaces import IHtmlConverter


class BsConverter(IHtmlConverter):

    def __init__(self, joiner=' '):
        self.joiner = joiner

    def to_text(self, html: Union[str, bytes], joiner=None, **kwargs) -> str:
        joiner = joiner if joiner is not None else self.joiner
        soup = BeautifulSoup(html, 'html.parser')

        # see https://stackoverflow.com/a/22800287/2667536
        # kill all script and style elements
        for script in soup(['script', 'style', 'form', 'code']):
            script.decompose()  # rip it out

        # strip all \n, which are sometimes used instead of spaces
        # - https://docs.streamlit.io/api.html?highlight=exception#streamlit.exception
        sents = (re.sub(' +', ' ', s.replace('\n', ' ')) for s in soup.stripped_strings)
        return joiner.join(sents)

    def __str__(self):
        return f'{self.__class__.__name__}({vars(self)})'.replace("'", '')

def main():
    import argparse, sys
    import requests
    parser = argparse.ArgumentParser()
    parser.add_argument('-u', '--url', nargs='+', default=[])
    parser.add_argument('-f', '--file', type=argparse.FileType('r'), default=None)
    parser.add_argument('-j', '--joiner', default=' ')
    args = parser.parse_args()

    jt = BsConverter(joiner=args.joiner)

    if args.file:
        print(f'=== from TEXT', file=sys.stderr)
        print(jt.to_text(args.file.read()))

    for url in args.url:
        try:
            response = requests.get(url)
            text = jt.to_text(response.content)
            print(f'=== from URL {url}', file=sys.stderr)
            print(text)
        except Exception as e:
            print(e)



#!/usr/bin/env python3

"""
This splitter implementation was heavily inspired from moses' split-sentences.perl. It requires only basic python 3 modules
and the `regex <https://pypi.org/project/regex/>`_ module (which supports unicode regexes).

Main changes compared to the latest split-sentences.perl:

* implemented with Swiss-German (Web) in mind: no support for Chinese, Hindi and Gujarati,
  numbers and quotes conventions based on this language only;
* addition of a "more" option that splits on ":;" characters (but tries to keep emojis and URLs intact);
* lowercase letters can also be signals of start-of-sentence (often seen on the web);
* more aggressive splits on ``?!`` characters;
* split returns a list of sentences instead of a string;
* it is possible to load multiple nonbreaking prefix lists (merged into one lookup table);
* no support for <p> tags (I didn't get the purpose of this option anyway)

"""

import os
import re
import sys

import regex
import logging

from ..interfaces import ISplitter

logger = logging.getLogger(__name__)

# Non-breaking prefix entries
_UNDEF = -1  #: the prefix doesn't exist
_ANY = 1  #: the prefix always applies
_NUMERIC_ONLY = 2  #: the prefix applies only when followed by number(s)


# Perl Regex substitutions:
# * \p{IsPi} => \p{Pi} or \p{Initial_Punctuation}: any kind of opening quote.
# * \p{IsPf} => \p{Pf} or \p{Final_Punctuation}: any kind of closing quote.


class MocySplitter(ISplitter):
    """
    A splitter largely inspired from Moses' split_sentences.perl. The name is a contraction of Moses and Lucy.
    Most important changes:

    * a start of sentence doesn't need to be an uppercase, lowercase letters will do too;
    * the more parameter let you choose to break on ';:' as well;
    * end of sentences '?!' are better handled
    * multiple nonbreaking_prefix files can be used at once (rules will be merged)

    """

    def __init__(self, langs=None, prefix_file=None, more=True, keep_newlines=True):
        """
        :param lang: a List[str] of language(s) for nonbreaking_prefix file to load (default: en, de)
        :param prefix_file: path to a custom nonbreaking_prefix file
        :param more: if set, systematically split on ``:;`` by default
        :param keep_newlines: if set, treat newlines as paragraph delimiters that will be preserved.
            If unset, newlines are ignored and empty lines are treated as paragraph delimiters
            (Moses original behavior, see :py:meth:`split`).

        """
        self.langs = langs if langs is not None else ['en', 'de']  #: nonbreaking prefixes files to load
        self.more = more  #: whether or not to split on ``:;``
        self.nb_prefixes = self.load_nb_prefixes(self.langs, prefix_file)  #: nonbreaking prefix lookup table
        self.keep_newlines = keep_newlines
        # TODO: keep_newlines seems better with Justext, but http://www.raere-waggis.ch/p300fasnacht2007.htm
        # things like songs etc ...

    def split(self, input_text, more=None, keep_newlines=None):  # -> List[str]
        """
        Split a text into sentences. Depending on the value of :py:attr:`.keep_newlines`, either :py:meth:`split_sentences`
        or :py:meth:`split_text` will be called.

        :param input_text: the input text
        :param more: override the class' parameter
        :param keep_newlines: override the class' parameter
        :return: a list of sentences (no blank lines)
        """
        more = more if more is not None else self.more
        keep_newlines = keep_newlines if keep_newlines is not None else self.keep_newlines
        return self._split_sentences(input_text, more) if keep_newlines else self._split_text(input_text, more)

    def _split_sentences(self, input_text, more):  # -> List[str]
        """
        Split a text into sentences. Newlines already present in text will be preserved and act as paragraph delimiters.

        :param input_text: the input text
        :param more: split on :; if true
        :return: a list of sentences (no blank lines)
        """
        return sum((
            self.split_paragraph(p, self.nb_prefixes, more)
            for p in input_text.split('\n')
            if p and not p.isspace()
        ), [])

    def _split_text(self, input_text, more):  # -> List[str]
        """
        Split a text into sentences. Newlines already present in text won't be preserved and empty lines act as
        paragraph delimiter.
        :param input_text: the input text
        :param more: split on :; if true
        :return: a list of sentences (no blank lines)
        """
        # equivalent of process_text in moses, but returns a list
        current_paragraph = ''
        splits = []
        for line in input_text.split('\n'):
            if not line or line.isspace():
                # Time to process this block; we've hit a blank or <p>
                if current_paragraph:
                    splits.extend(self.split_paragraph(current_paragraph, self.nb_prefixes, more))
                    current_paragraph = ""
            else:
                current_paragraph += line + ' '

        if current_paragraph:
            # Do the leftover text.
            splits.extend(self.split_paragraph(current_paragraph, self.nb_prefixes, more))

        return splits

    @classmethod
    def cleanup_spaces(cls, text):  # -> str
        """Normalize spaces in a text."""
        # clean up spaces
        text = re.sub(' +', ' ', text)
        text = re.sub('\n ', '\n', text)
        text = re.sub(' \n ', '\n', text)
        return text.strip()

    @classmethod
    def split_paragraph(cls, text, nb_prefixes, more=False):  # -> List[str]
        """
        Handle one paragraph of text.

        :param text: the paragraph to split
        :param nb_prefixes: the dictionary of nonbreaking_prefix (see perl implementation/doc)
        :param more: if set, systematically split on :;
        :return: a list of sentences
        """
        # text is one paragraph, this is equivalent to the preprocess perl method.
        if not text:
            return ''
        # clean up spaces
        text = cls.cleanup_spaces(text)

        ##### Add sentence breaks as needed #####
        if more:
            # this one is present in the python wrapper, see
            # https://bitbucket.org/luismsgomes/mosestokenizer/src/default/src/mosestokenizer/split-sentences.perl
            # text = regex.sub(r'([\:;])', r'\1\n', text)
            # TODO: improvement: try to keep emojis, numers like 1:1 and urls intact
            text = regex.sub(r'([\:;])([^\d\)\(/-])', r'\1\n\2', text)

        # split if ?! is followed by a lowercase (often on the web)
        text = regex.sub(r'([\?!]+)([^\?!\p{Pe}\p{Pf}\"])', r'\1\n\2', text)
        # text = regex.sub(r'([?!]) +([\'\"\(\[\¿\¡\p{Pi}]*[\p{L}])', r'\1\n\2', text)

        # Multi-dots followed by sentence starters.
        text = regex.sub(r'(\.[\.]+) +([\'\"\(\[\¿\¡\p{Pi}]*[\p{L}])', r'\1\n\2', text)

        # Add breaks for sentences that end with some sort of punctuation
        # inside a quote or parenthetical and are followed by a possible
        # sentence starter punctuation and ~upper case~ letter
        text = regex.sub(r'([?!\.][\ ]*[\'\"\)\]\p{Pf}]+) +([\'\"\(\[\¿\¡\p{Pi}]*[\ ]*[\p{Lu}])', r'\1\n\2', text)

        # Add breaks for sentences that end with some sort of punctuation,
        # and are followed by a sentence starter punctuation and upper case letter.
        text = regex.sub(r'([?!\.]) +([\'\"\(\[\¿\¡\p{Pi}]+[\ ]*[\p{L}])', r'\1\n\2', text)

        # Special punctuation cases are covered. Check all remaining periods.
        words = text.split(' ')
        text = ''
        for i in range(len(words) - 1):
            # TODO: add the # as a possible sentence start ? (twitter and hashtags)
            m = regex.search(r'([\p{IsAlnum}\.\-]*)([\'\"\)\]\%\p{Pf}]*)(\.+)$', words[i])
            if m is not None:
                # Check if $1 is a known honorific and $2 is empty, never break.
                prefix, starting_punct, _ = m.groups()
                if prefix and nb_prefixes.get(prefix, _UNDEF) == _ANY and not starting_punct:
                    pass  # Not breaking prefix
                elif regex.search(r'(\.)[\p{IsUpper}\-]+(\.+)$', words[i]) is not None:
                    pass  # Not breaking - upper case acronym
                elif regex.search(r'^([ ]*[\'\"\(\[\¿\¡\p{Pi}]*[ ]*[\p{L}0-9])', words[i + 1]):
                    # The next word has maybe a bunch of initial quotes, maybe a
                    # space, then either ~upper case~ letter or a number
                    if prefix and nb_prefixes.get(prefix, _UNDEF) == _NUMERIC_ONLY and not starting_punct \
                            and regex.search('^[0-9]+', words[i + 1]):
                        # exception: we have a numeric-only prefix followed by a number
                        pass
                    else:
                        # In any other case, split
                        words[i] = words[i] + '\n'
            text += words[i] + ' '

        # We stopped one token from the end to allow for easy look-ahead.
        # Append it now.
        text = text + words[-1]

        # clean up spaces
        text = cls.cleanup_spaces(text)

        return text.split('\n')

    @classmethod
    def load_nb_prefixes(cls, langs, prefix_file=None):  # -> Dict
        """
        Read the nonbreaking_prefixes from a file or from an array of languages.

        :param langs: the language(s) to load
        :param prefix_file: a custom file to load (has priority over lang)
        :return: the nonbreaking_prefix dictionary, with key=prefix and value=1|2.
                1 means apply anywhere, 2 means only applies when followed by digits.
        """
        if prefix_file is not None:
            print(f'Using custom prefix file: {prefix_file}', file=sys.stderr)
            return cls._read_prefix_file(prefix_file)

        this_dir = os.path.dirname(os.path.realpath(__file__))
        prefix_file_pattern = os.path.join(this_dir, 'moses_splitter_prefixes.{}.txt')

        prefixes = dict()
        for lang in langs:
            prefix_file = prefix_file_pattern.format(lang)
            if not os.path.isfile(prefix_file):
                logger.warning(f'No known abbreviations for language {lang}, skipping...')
                continue
            prefixes.update(**cls._read_prefix_file(prefix_file))
            logger.info(f'Loaded prefix file for lang={lang} (prefix size: {len(prefixes)}).')
        return prefixes

    @staticmethod
    def _read_prefix_file(filename):  # -> Dict
        # load a nonbreaking_prefix lookup table from a file
        prefixes = dict()
        with open(filename) as f:
            for line in f:
                line = line.strip()
                if len(line) == 0 or line.startswith('#'):
                    continue
                if line.endswith('#NUMERIC_ONLY#'):
                    prefixes[line.replace('#NUMERIC_ONLY#', '').strip()] = _NUMERIC_ONLY
                else:
                    prefixes[line] = _ANY
        return prefixes


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', type=argparse.FileType('r'), default='-')
    parser.add_argument('-o', '--out', type=argparse.FileType('w'), default='-')
    parser.add_argument('-l', '--lang', action='append')
    parser.add_argument('-pf', '--prefix-file', default=None)
    parser.add_argument('-m', '--more', default=False, action='store_true')

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO, stream=sys.stderr, format='%(levelname)s: %(msg)s')

    splitter = MocySplitter(
        langs=args.lang,
        prefix_file=args.prefix_file,
        more=args.more)

    args.out.write('\n'.join(
        splitter.split(args.input.read())
    ))



#!/usr/bin/env python3

#
# A python script that mimics what Moses remove-non-printing-char.perl
# and normalize-punctuation.perl do, with some slight improvements:
#  - normalize combining diacritics at the beginning (and remove any extra one, e.g. Gguppò̃ò̃s -> Gguppòòs)
#  - properly handle GSW apostrophes (i.e. ' surrounded by accented/special chars)
#  - replace [non-breakable spaces+space separator unicode category]
#    by regular spaces as a last step
#
# Lucy Linder, June 2019
#

__all__ = ['Normalizer', 'normalize_text']

import argparse
import re
import sys
import unicodedata

REG, STR = 0, 1  # flags for using re.sub vs string.replace

# from Leipzig rules
wrong_encoding_pattern = re.compile(r'\u0007|\u007F|\u0080-\u00A0|\u00C2|\u00C3|\u0084\uFFFD|\uF0B7|\u00AD', re.UNICODE)
# misc symbols + http://stackoverflow.com/a/13752628/6762004
emoji_pattern = re.compile('[\u2300-\u23ff\u2b50-\u2b55\u2600-\u2800\U00010000-\U0010FFFF]', re.UNICODE)

spaces_pattern = re.compile(
    # non-breakable space + https://www.compart.com/en/unicode/category/Zs
    '[\u00A0\u1680\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200A\u202F\u205F\u3000 ]+',
    flags=re.UNICODE
)

normalization_patterns = [
    (t, r if t == STR else re.compile(r), s) for (t, r, s) in
    [  # largely inspired from Moses normalize-punctuation.perl (lang=de)
        # strip control chars and \t, \r, but not \n (0x0A) + \u200d (zero-width joiner)
        (REG, u'[\x00-\x09\x0B-\x1F\x7F-\x9F\u200d]', ' '),
        # strip soft-hyphen (and don't replace it by space: http://jkorpela.fi/shy.html)
        (STR, '\u00AD', ''),
        # strip extra combining diacritics (given unicodedata.normalize was run prior to this)
        (REG, '[\u0300-\u036F\uFE00-\uFE0F]', ''),
        # replace variation selectors (0xFE0F is often used, sometimes in a row...)
        (REG, '[\u0300-\u036F\uFE00-\uFE0F]', ' '),
        # strip the � character, except when it might help detect a wrong encoding issue (shouldn't happen if ftfy is installed)
        (REG, r'([^\u0084]?)\uFFFD+', r'\1'),
        # normalize unicode punctuation
        (STR, '`', "'"),
        (STR, '„', '"'),
        (STR, '“', '"'),
        (STR, '”', '"'),
        (STR, '—', ' - '),
        (REG, u'[\u00AF\u2010-\u2015\u2212\uFE58\uFE63\uFF0D\u1806]', '-'),  # dashes
        (STR, '´', "'"),
        (REG, r'([^\W\d_])[‘’]([^\W\d_])', r"\1'\2"),  # I
        (STR, '‘', '"'),
        (STR, '’', '"'),
        (STR, '‛', '"'),
        (STR, u'\u0092', "'"),
        (STR, u'\u0093', '"'),
        (STR, '‚', '"'),
        (STR, "''", '"'),
        (STR, '…', '...'),
        # French quotes
        (STR, '\u00A0«\u00A0', ' "'),
        (STR, '«\u00A0', '"'),
        (STR, '«', '"'),
        (STR, '\u00A0»\u00A0', '" '),
        (STR, '\u00A0»', '"'),
        (STR, '»', '"'),
        # other symbols
        (STR, '‹', '<'),
        (STR, '›', '>'),
        # ligatures
        (STR, 'œ', 'oe'),
        (STR, 'æ', 'ae'),
        (STR, 'ﬁ', 'fi'),
        (STR, 'ﬀ', 'ff'),
        (STR, 'ﬂ', 'fl'),
        (STR, 'ĳ', 'ij'),
        # remove pseudo-spaces in specific settings
        (STR, '\u00A0%', '%'),
        (STR, '\u00A0:', ':'),
        (STR, '\u00A0?', '?'),
        (STR, '\u00A0!', '!'),
        (STR, '\u00A0;', ';'),
        # numbers
        (REG, r'(\d)[\u00A0,](\d)', r'\1\2'),  # remove non-breakable spaces or "," in numbers
        # German/Spanish/French "quotation", followed by comma, style
        (REG, r'(\.+)"(\s*[^<])', r'"\1\2'),  # don't fix period at end of sentence TODO: what ??
        (STR, ',"', '",'),
        # ensure , is not left alone
        (STR, ' ,', ','),
        (STR, ',', ', '),
        # normalize space
        (REG, spaces_pattern, ' '),
        # normalize spaces
        (REG, r'(\d) \%', r'\1%'),
        # the following is a very bad idea because of emojis
        # (STR, '(', ' ('),
        # (STR, ')', ') '),
        # (REG, r'\) +([\.\!\:\?\;\,])', r')\1'),
        # (STR, '( ', '('),
        # (STR, ' )', ')'),
        # (STR, ' :', ':'),
        # (STR, ' ;', ';'),
        # normalize spaces around, trying to avoid emojis and numbers, e.g. (arXiv:133)
        (REG, r'([\w"\']) ?(:|;) ?([^\W\d]|["\'\n]|$)', r'\1\2 \3'),
        (STR, ' , ', ', '),
        (STR, ' .', '.'),  # TODO: useful ?
        (REG, r'\( +(\w|\d)', r'(\1'),
        (REG, r'(\w|\d) +\)', r'\1)'),
    ]]


def normalize_text(text, fix_encoding=False, strip_emojis=False):
    """
    Normalize text:

    * normalize accents (using NFC convention),
    * strip control/invisible chars and leftover combining diacritics,
    * undo ligatures,
    * normalize quotes, apostrophes and unicode characters (dashes, etc.),
    * normalize spaces (all spaces, including nbsp and tabs, will be encoded as 0x20),
    * strip and collapse multiple spaces into one,
    * etc.

    Optionally:

    * try to detect and fix encoding issues (see `ftfy.fix_encoding <https://ftfy.readthedocs.io/en/latest/>`_)
       on a per-sentence basis (delimited by newlines);
    * strip emojis (using a simple regex, not all cases are covered !)


    :param text: the text to normalize, newlines will be preserved;
    :param fix_encoding: if set, use ftfy to fix encoding issues on a per-sentence basis;
    :param strip_emojis: if set, try to find and strip unicode emojis;
    :return: the normalized text
    """
    # optionally fix encoding using ftfy
    if fix_encoding and wrong_encoding_pattern.search(text) is not None:
        try:
            import ftfy
            text = '\n'.join(
                ftfy.fix_encoding(t) if wrong_encoding_pattern.search(t) is not None else t
                for t in text.split('\n')
            )
        except ModuleNotFoundError:
            print('WARNING: norm_punc.py, fixing encoding requires the ftfy package: pip install ftfy.')

    # normalize (e.g. combining diacritics)
    text = unicodedata.normalize('NFC', text)

    # optionally strip emojis
    if strip_emojis:
        # I formally used the emoji library, which is really nice but slooooow (and doesn't cover ASCII misc symbols).
        # I thus preferred to use a simpler regex that covers most cases is is waaaay faster
        # (203ms to process 164343 short sentences, against 31s with emoji)
        text = emoji_pattern.sub(' ', text)

    # apply patterns in order
    for i, (typ, pattern, replace) in enumerate(normalization_patterns):
        if typ == REG:
            text = pattern.sub(replace, text)
        else:
            text = text.replace(pattern, replace)

    # normalize spaces
    text = spaces_pattern.sub(' ', text)

    # don't forget to normalise spaces in the beginning and end
    text = re.sub(r'(^|\n)\s+', r'\1', text)
    text = re.sub(r'\s+(\n|$)', r'\1', text)

    return text


class Normalizer():
    """A wrapper around :py:meth:`normalize_text`"""

    def __init__(self, **kwargs):
        """Initialize a normalizer. The ``kwargs`` will be passed to :py:meth:`normalize_text` as-is. """
        self.kwargs = kwargs  #: extra options to pass to :py:meth:`normalize_text`

    def normalize(self, text, **kwargs):
        """Call :py:meth:`normalize_text` on ``text`` with the extra :py:attr:`kwargs` options."""
        options = self.kwargs
        if kwargs:
            options = dict(**self.kwargs)
            options.update(kwargs)
        return normalize_text(text, **options)


# ---

def main():
    parser = argparse.ArgumentParser(
        description=
        "Normalise text (remove control chars, uncurl quotes, normalise spaces...)"
    )
    parser.add_argument('-i', type=argparse.FileType('r'), required=True)
    parser.add_argument('-fe', '--fix-encoding', default=False, action='store_true')
    parser.add_argument('-se', '--strip-emojis', default=False, action='store_true')
    parser.add_argument('-o', '--out', type=argparse.FileType('w'), default=sys.stdout)
    args = parser.parse_args()

    try:
        text = normalize_text(args.i.read(), fix_encoding=args.fix_encoding, strip_emojis=args.strip_emojis)
        args.out.write(text)
    except ModuleNotFoundError as e:
        print(e)
        exit(1)



"""
This module relies on `JusText <https://github.com/miso-belica/jusText>`_ for text extraction.

To get a feel, try the `online demo of the original JustText <http://nlp.fi.muni.cz/projects/justext/>`_.

.. note::

    * usually, jusText uses ``''`` (empty string) to join text nodes inside a paragraph, thus making things like
      "*One sentence.Second sentence*" likely. Here, we always use a space to join, then normalize the spaces.
    * justext will throw an error on an empty document content.

"""
import logging
import re
from typing import Union

from bs4 import UnicodeDammit
import justext
from ..interfaces import IHtmlConverter

logger = logging.getLogger(__name__)


class JustextConverter(IHtmlConverter):
    """
    An HTML converter that relies on
    `JusText <https://github.com/miso-belica/jusText>`_ to cleverly extract meaningful text from webpages.
    """

    def __init__(self, joiner='\n',
                 keep_bad=True,
                 stoplist=None,
                 stopwords_low=justext.core.STOPWORDS_LOW_DEFAULT,
                 stopwords_high=justext.core.STOPWORDS_HIGH_DEFAULT,
                 **kwargs):
        """
        Create a crawler instance.
        :param joiner: character used to join paragraphs;
        :param keep_bad: if set, keep everything.
        If unset, keep only paragraphs with a context-free class of "neargood" or "good".
        :param stoplist: see the `justText doc <https://github.com/miso-belica/jusText/blob/dev/doc/algorithm.rst>`_
        :param stopwords_low: idem
        :param stopwords_high: idem
        :param kwargs: unused
        """
        self.joiner = joiner

        if stoplist is not None:
            with open(stoplist) as f:
                self.kwargs = dict(
                    stoplist=set(l.strip() for l in f if len(l)),
                    stopwords_low=stopwords_low,
                    stopwords_high=stopwords_high)
        else:
            self.kwargs = dict(stoplist=set(), stopwords_low=0, stopwords_high=0)

        self.kwargs.update(kwargs)
        self.keep_bad = keep_bad
        logger.debug(self)

    def to_text(self, html: Union[str, bytes], joiner=None, keep_bad=None, **kwargs) -> str:
        if keep_bad is None:
            keep_bad = self.keep_bad
        if joiner is None:
            joiner = self.joiner

        try:
            # justext uses the decode/replace strategy by default, so encoding errors shouldn't happen
            # see justext core.py:DEFAULT_ENC_ERRORS ... well I don't know why, but errors are still triggered,
            # but not if I do the decoding here... maybe the encoding parameter ? Problematic URLs examples
            # - http://www.triumphowners.ch/index.php?page=Thread%3D654%3D13
            # - https://angerweit.tikon.ch/lieder/lied.php?src=folk-de%2Fsimeli
            if isinstance(html, bytes):
                dammit = UnicodeDammit(html, [None], is_html=True) # way more precise than requests
                if 'encoding' in kwargs and dammit.original_encoding != kwargs['encoding']:
                    logger.debug(f"encoding: declared={kwargs['encoding']}, dammit={dammit.original_encoding}")
                html = html.decode(encoding=dammit.original_encoding, errors='replace')
            paragraphs = justext.justext(html, **self.kwargs)
            # paragraphs = justext.justext(content, encoding=soup.original_encoding, **self.kwargs)
            text_blocks = (self._get_text(p) for p in paragraphs if keep_bad or 'good' in p.cf_class)
            return joiner.join(text_blocks)
        except Exception as e:
            if 'Document is empty' in str(e):
                return ''  # Document is empty
            else:
                raise e

    def _get_text(self, p):
        # mmhh... In justext, they join on '', such that we often have things like:
        # "end of sentence.Start of sentence".
        text = ' '.join(p.text_nodes).replace('\n', ' ')
        return re.sub(' +', ' ', text).strip()

    def __str__(self):
        return f'{self.__class__.__name__}({vars(self)})'.replace("'", '')


def main():
    import argparse, sys
    import requests
    parser = argparse.ArgumentParser()
    parser.add_argument('-u', '--url', nargs='+', default=[])
    parser.add_argument('-f', '--file', type=argparse.FileType('r'), default=None)
    parser.add_argument('-j', '--joiner', default='\n')
    parser.add_argument('-b', '--keep-bad', default=False, action='store_true',
                        help='Return all found sentences (vs "good" only).')
    args = parser.parse_args()

    jt = JustextConverter(joiner=args.joiner, keep_bad=args.keep_bad)

    if args.file:
        print(f'=== from TEXT', file=sys.stderr)
        print(jt.to_text(args.file.read()))

    for url in args.url:
        try:
            response = requests.get(url)
            text = jt.to_text(response.content)
            print(f'=== from URL {url}', file=sys.stderr)
            print(text)
        except Exception as e:
            print(e)



"""
A simple utility module module to deal with links. Given a list of links and a base url, it will:

* resolve relative links,
* exclude duplicates,
* strip anchors,
* exclude non HTTP/HTTPs links.

To further filter links pointing to images, PDFs or
other non-text resources, look at the `is_media_link` method (also provided).
"""

from typing import Iterable, Generator, Optional, Union
import urllib.parse as up


MEDIA_EXTENSIONS = dict((s, True) for s in [
    "3dv", "3g2", "3gp", "pi1", "pi2", "pi3", "ai", "amf", "amv", "art", "art", "ase", "asf", "avi", "awg", "blp",
    "bmp", "bw", "bw", "cd5", "cdr", "cgm", "cit", "cmx", "cpt", "cr2", "cur", "cut", "dds", "dib", "djvu", "doc",
    "docx", "drc", "dxf", "e2d", "ecw", "egt", "egt", "emf", "eps", "exif", "f4a", "f4b", "f4p", "f4v", "flv", "flv",
    "flv", "fs", "gbr", "gif", "gif", "gifv", "gpl", "grf", "hdp", "icns", "ico", "iff", "iff", "int", "int", "inta",
    "jfif", "jng", "jp2", "jpeg", "jpg", "jps", "jxr", "lbm", "lbm", "liff", "m2v", "m4p", "m4v", "m4v", "max", "miff",
    "mkv", "mng", "mng", "mov", "mp2", "mp4", "mpe", "mpeg", "mpeg", "mpg", "mpg", "mpv", "msp", "mxf", "nitf", "nrrd",
    "nsv", "odg", "ogg", "ogv", "ota", "pam", "pbm", "pc1", "pc2", "pc3", "pcf", "pct", "pcx", "pcx", "pdd", "pdf",
    "pdn", "pgf", "pgm", "pict", "png", "pnm", "pns", "ppm", "ppt", "pptx", "psb", "psd", "psp", "px", "pxm", "pxr",
    "qfx", "qt", "ras", "raw", "rgb", "rgb", "rgba", "rle", "rm", "rmvb", "roq", "sct", "sgi", "sgi", "sid", "stl",
    "sun", "svg", "svi", "sxd", "tga", "tga", "tif", "tiff", "v2d", "vnd", "vob", "vrml", "vtf", "wdp", "webm", "webp",
    "wmf", "wmv", "x3d", "xar", "xbm", "xcf", "xls", "xlsx", "xpm", "yuv"
])

def process_links(base_url: Optional[str], links: Iterable[str]) -> Generator[str, None, None]:
    """
    Resolve, clean and filter links found in a page. By links we mean here any value of `href` attribute found
    in a page. This is especially useful for the

    In addition to what :py:meth:`fix_url` does, this method will also exclude:

    * the base URL
    * Duplicates

    For two links to be considered duplicates, they need to match exactly. Exceptions are anchors (stripped
    automatically) and trailing slashes (in this case, the first encountered link will be returned).

    :param base_url: the URL of the current page (not returned and used to resolve relative links). Use '' to ignore.
    :param links: a list of links found, relative or absolute
    :return: a generator of unique absolute URLs, all beginning with `http`
    """
    seen = set()  # keep a set of seen urls

    if base_url is not None:
        seen.update([base_url, base_url + '/'])

    for link in links:
        fixed_url, ok = fix_url(link, base_url)
        if ok and fixed_url not in seen:
            yield fixed_url
            # add it as is
            seen.add(fixed_url)
            # avoid duplicate links with just an ending slash that differ
            seen.add(fixed_url[:-1] if fixed_url.endswith('/') else fixed_url + '/')


def fix_url(url: str, base_url: str = None) -> (str, bool):
    """
    Fix/normalize an URL and decide if it is interesting to crawl.

    The URL is potentially transformed by:

        * resolving relative to absolute URLs (only if base_url is set)
        * removing anchors

    The "is interesting" decision will exclude:

        * relative URLs (so ensure to provide a base_url if the url is relative)
        * non HTTP links (`mailto:`, `javascript:`, anchors, ...)

    :param url: the url
    :param base_url: the base url, required if the url is a relative one
    :return: a tuple (fixed_url, is_interesting)
    """
    if base_url is not None:
        # make relative into absolute links
        if base_url.endswith('/'):
            # remove the ending "/", because of a weird behavior of urljoin:
            #  urljoin('http://example.com/page1/', 'page2') => 'http://example.com/page1/page2'
            #  urljoin('http://example.com/page1', 'page2') => 'http://example.com/page2'
            base_url = base_url[:-1]
        url = up.urljoin(base_url, url)

    parsed: up.ParseResult = up.urlparse(url)

    if parsed.fragment:
        parsed = parsed._replace(fragment='')

    # ensure it is a HTTP or HTTPS URL
    if not (parsed.scheme and parsed.scheme.startswith('http')):
        return up.urlunparse(parsed), False  # return what we got so far

    # we got this far, the URL should not be fixed and worth keeping
    return up.urlunparse(parsed), True


def is_media_link(url: Union[up.ParseResult, str]) -> bool:
    """
    Check if an url contains a known media extension
    :param url: a URL, either in form of a string or an already "parsed" URL
    :return: true if most probably a media link (pdf, image, video, audio, etc.) false otherwise
    """
    if isinstance(url, str):
        parsed: up.ParseResult = up.urlparse(url)
    elif isinstance(url, up.ParseResult):
        parsed = url
    else:
        raise Exception(f'Expecting a string or a urlparse.ParseResult, got {type(url)}')

    # path contains a media extension
    # or query ends with a media extension (e.g. ?doc=lala.pdf)
    return '.' in parsed.path and parsed.path.split(".")[-1].lower() in MEDIA_EXTENSIONS or \
           '.' in parsed.query and parsed.query.split(".")[-1].lower() in MEDIA_EXTENSIONS



# suppress warning for invalid SSL certificates
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

from .html_converters import *
from .filterers import *
from .splitters import *
from .norm_punc import Normalizer, normalize_text

from .interfaces import IHtmlConverter, ISplitter, INormalizer, IFilterer

from .links import *



import phrasal
from phrasal import *

if __name__ == '__main__':

    import sys
    import inspect

    if len(sys.argv) < 2:
        print('Missing requirement argument: <ClassName>')
        sys.exit(1)

    clsname = sys.argv[1]
    if clsname in ['-h', '--help', 'help', '?']:
        print('Call one of the tools from the commandline. Usage: ')
        print('   classname [other arguments specific to classname]|[-h]')
        print('\nAllowed classname arguments:')
        for k, v in phrasal.__dict__.items():
            if k[0].isupper():
                if getattr(inspect.getmodule(v), 'main', None) is not None:
                    print(' -', v.__name__)
        exit(0)

    if clsname not in globals():
        print(f'Class "{clsname}" could not be loaded. Maybe a typo or error in casing ?')
        exit(1)

    cls = globals()[clsname]
    main_func = getattr(inspect.getmodule(cls), 'main', None)
    if main_func is None:
        print(f'Class "{clsname}" does not define a main() method.')
        exit(1)

    sys.argv.pop(0)
    main_func()



from typing import List

from bs4 import BeautifulSoup
from get_html.env_defined_get import do_get

from . import link_utils


def extract_links(url: str) -> List[str]:
    resp = do_get(url)
    soup = BeautifulSoup(resp.content, 'html.parser')

    # see https://stackoverflow.com/a/22800287/2667536
    # kill all script and style elements
    for script in soup(['script', 'style', 'form']):
        script.decompose()  # rip it out

    links = (a.get('href') for a in soup.find_all('a', href=True))
    return [l for l in link_utils.process_links(url, links)]


import setuptools
import os

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

setuptools.setup(
    name='phrasal',
    version='0.0.1',
    author='Lucy Linder',
    author_email='lucy.derlin@gmail.com',
    license='Apache License 2.0',
    description='NLP tools to extract, normalize and filter sentences from text/HTML',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/derlin/phrasal',

    # install all packages found under src/
    package_dir={'': 'src'},
    packages=setuptools.find_packages('src', exclude=['showcase']),

    # include other files such as html, css, etc
    include_package_data=True,  # read from MANIFEST.in
    zip_safe=True,  # not safe if the library needs to read package data from os (path is different from a zip)

    python_requires='>=3.6',

    classifiers=(
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ),

    # regular dependencies
    install_requires=[
        'requests>=2.22,<2.23',
        'regex>=2020.1, <2020.2',
        'bs4',
        'justext==2.2.0',
        'ftfy>=5.6',
        'pyyaml>=5.3',
        'get-html',
    ],
    # extra dependencies
    extras_require={
        'showcase': ['streamlit']
    }
)



"""
This module contains an implementation of :py:class:`~swisstext.cmd.scraping.interfaces.ISentenceFilter` that uses
simple rules to filter "well-formed" sentences.

How it works
------------
Each sentence is checked against a list of rules and rejected / considered invalid if any of those rules fail.
Rules are thus *AND-based*.

Rules are defined using a simple YAML syntax and can be of two types: *length-based* (character count)
 or *pattern-based* (regular expressions). They are checked in the same order they are defined.

.. note::

    * Regular expressions can be quite expensive, so try to limit their complexity to the minimum required.
    * Rules are checked in the same order as they are defined, so it is advised to put the most generic / efficient
      ones first.

.. note::

    This module uses the `regex library <https://pypi.org/project/regex/>`_ (version V0)
    instead of the default re. You can thus freely use
    `unicode regular expressions <https://www.regular-expressions.info/unicode.html>`_
    in your rules.

Rule syntax
-----------

**Length-based rules (length)** must specify *at least one* of *min* or *max* length,
i.e. the bounds on the number of characters.
The rule succeeds if ``min <= len(s) <= max``. Here is an example:

.. code-block:: yaml

    - max_length:
      descr: too long
      length:
        max: 1000

**Pattern-based rules (find)** a bit similar, but instead of counting the number of characters, they count the number
of occurrences of a *pattern* (i.e. the number of matches when calling ``regex.findall(pattern, s)``).
The rule succeeds if ``min <= nb_matches <= max`` (inclusive !)


Examples:

.. code-block:: yaml

    - dashes:
      descr: too many dashes
      find:
        pattern: '[\u00AF\u2010\u2015\u2212\uFE58\uFF0D-]'
        count:
          max: 2

    - right_number_of_punctuation:
      descr: punctuation is between 5 and 10
      find:
        pattern: '\p{P}'
        count:
          min: 5
          max: 10

**Comparison rules (compare)** they define both a numerator and a denominator pattern. The number of matches is
found for each pattern, then a ratio is computed as:
``ratio = count(num matches) / (count(denom matches)+1)``. The matches is once again based on ``regex.findall``.
The rule succeeds if ``min <= ratio <= max`` (inclusive !).

Example:

.. code-block:: yaml

    - too_many_commas:
      descr: compare the number of commas against the number of words in the sentence
      compare:
        num: ','
        denom: '\p{L}+'
        ratio:
          max: 0.25

Finally, **an if condition** can be used. If conditions are checked first, and if the check fails, the rule is
simply ignored:

.. code-block:: yaml

    - ellipsis:
      descr: ellispsis on short sentences.
      if:
        length:
          max: 30
      find:
        pattern: '(\.\s?){3}$'
        count:
          max: 0

Rules can additionnally specify ``examples`` and ``counterexamples`` that could be used to check quickly if they work
(see the ``Rule.self_check`` method). For example:

.. code-block:: yaml

    - spelled_words:
      descr: W O R D
      find:
        pattern: ' ([\p{L}] ){3,}'
        count:
          max: 0
      examples:
        - 'you must B E L I E V E me.'
        - 'span spells S P A N.'
        - 's p a n means span!!'
      counterexamples:
        - 'this is O K :)'

"""

import regex
import yaml
import logging
from os import path

from ..interfaces import IFilterer

logger = logging.getLogger(__name__)


# TODO: a good way to detect encoding errors is to compare the result of
# len(re.findall('[^\W\d]')) and len(regex.findall('\p{L}'))

class PatternSentenceFilter(IFilterer):
    """
    By default, rules are loaded from the default file ``pattern_sentence_filter.yaml`` in the current directory.
    You can override this by passing a path to the constructor (``rulespath`` argument).
    """

    def __init__(self, rulespath=None):
        """Load rules from the default YAML file or the path provided."""
        if rulespath is None:
            rulespath = path.join(path.dirname(path.realpath(__file__)), 'pattern_sentence_filter.yaml')

        self.rulespath = rulespath
        self.rules = Rules(yaml.safe_load(open(rulespath)))

    def is_valid(self, sentence):
        """Returns true only if all the rules were respected."""
        return not self.rules.is_invalid(sentence)


class MinMax:
    """Encapsulates and handles min/max bounds. A bound set to -1 will be ignored."""

    def __init__(self, min=-1, max=-1):
        self.min = min
        self.max = max

    def is_invalid(self, s) -> bool:
        return self.is_out_of_range(len(s))

    def is_out_of_range(self, n) -> bool:
        return (self.min >= 0 and self.min > n) or (self.max >= 0 and self.max < n)

    def __repr__(self):
        return "(min={}, max={})".format(self.min, self.max)


class Compare:
    def __init__(self, num, denom, ratio):
        self.num = regex.compile(num)
        self.denom = regex.compile(denom)
        self.ratio = MinMax(**ratio)

    def is_invalid(self, s):
        ratio = len(self.num.findall(s)) / (len(self.denom.findall(s)) + 1)
        return self.ratio.is_out_of_range(ratio)

    def __repr__(self):
        return "Compare(num=%s, denom=%s, ratio=%s)" % (self.num, self.denom, self.ratio)


class Find:
    """Handles pattern-based rule logic (find entry in yaml)"""

    def __init__(self, pattern, count=None, ratio=None):
        if count is None and ratio is None:
            logger.warning(f"{pattern}: missing find condition: count or ratio...")
        self.pattern = regex.compile(pattern)
        self.count = MinMax(**count) if count else None
        self.ratio = MinMax(**ratio) if ratio else None

    def is_invalid(self, s):
        matches = self.pattern.findall(s)
        nb_matches = len(matches)
        if self.count and self.count.is_out_of_range(nb_matches):
            return True
        if self.ratio:
            ratio = nb_matches / len(s)
            return self.ratio.is_out_of_range(ratio)
        return False

    def __repr__(self):
        return "Find(pattern=%s, count=%s, ratio=%s)" % (self.pattern, self.count, self.ratio)


class Rule:
    """Encapsulates one rule"""

    def __init__(self, id, descr, find=None, compare=None, length=None, examples=None, counterexamples=None, **kwargs):
        self.id = id
        self.descr = descr
        self.examples = examples
        self.counterexamples = counterexamples
        self.iff = []
        # TODO: better way ?
        if 'if' in kwargs:  # if is a reserved keyword in python
            if 'length' in kwargs['if']:
                self.iff.append(MinMax(**kwargs['if']['length']))
            if 'pattern' in kwargs['if']:
                self.iff.append(Find(**kwargs['if']['pattern']))

        if length is not None:
            self.logic = MinMax(**length) if length else None
        elif find is not None:
            self.logic = Find(**find)
        elif compare is not None:
            self.logic = Compare(**compare)
        else:
            raise Exception('Found a rule with no length, find or ratio defined.')

    def is_applicable(self, s) -> bool:
        """Check for the if condition"""
        return not any([iff.is_invalid(s) for iff in self.iff])

    def is_invalid(self, s) -> bool:
        if self.is_applicable(s):
            if self.logic.is_invalid(s):
                logger.debug("%s FAILED on |%s|" % (self, s))
                return True
            return False
        else:
            # logger.debug("SKIPPED   RULE %s: |%s|" % (self.descr, s))
            return False

    def self_check(self, verbose=True) -> bool:
        passed = True
        for examples, expected in [(self.examples, True), (self.counterexamples, False)]:
            if examples is not None:
                if verbose: print(f' Checking {len(examples)} {str(expected):5s} examples...', end=' ', flush=True)
                fails = [e for e in examples if not (self.is_invalid(e) == expected)]
                if len(fails):
                    if verbose: print('Failed on:\n   ', '\n   '.join(fails))
                    passed = False
                else:
                    if verbose: print('OK.')
        return passed

    def __str__(self):
        return "RULE %d %s" % (self.id, self.descr)

    def __repr__(self):
        return "{}: [if {}] logic={}".format(self.descr, self.iff, self.logic)


class Rules:
    """
    This class represents a list of rules.
    """

    def __init__(self, rules_dict):
        """
        :param rules_dict: a dictionary of rules (as loaded by yaml)
        """
        self.rules = [Rule(idx + 1, **r) for (idx, r) in enumerate(rules_dict)]  # [:1]

    def is_invalid(self, sentence: str) -> bool:
        """Returns true if any rule that apply failed."""
        for idx, r in enumerate(self.rules):
            if r.is_invalid(sentence):
                # print("RULE %d %s FAILED on |%s|" % (idx, r.descr, sentence))
                return True
        return False

    def print_rules(self):
        """Prints all the rules, useful for debug."""
        for idx, r in enumerate(self.rules):
            print(idx, "=>", r.__repr__())

    def __getitem__(self, idx):
        return self.rules[idx]

    def __len__(self):
        return len(self.rules)


def main():
    import argparse
    import sys

    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', type=argparse.FileType('r'), default='-')
    parser.add_argument('-o', '--out', type=argparse.FileType('w'), default='-')
    parser.add_argument('-r', '--rules-file', default=None)

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO, stream=sys.stderr, format='%(levelname)s: %(msg)s')

    psf = PatternSentenceFilter(rulespath=args.rules_file)

    args.out.write('\n'.join(
        t for t in args.input if psf.is_valid(t)
    ))
